use crate::{model::DataPoint, settings::ConfigFile};
use std::{path::Path, sync::Arc};
use tch::Tensor;

pub struct DataLoader {
    pub batch_size: i64,
    pub len: usize,
    pub inputs: Vec<Tensor>,
    pub labels: Vec<Tensor>,
    pub pos: usize,
}

impl DataLoader {
    pub fn new(
        config: Arc<ConfigFile>,
        data: Vec<DataPoint>,
        classes_len: i64,
        batch_size: i64,
    ) -> DataLoader {
        let len: f64 = (data.len() as f64) / (batch_size as f64);
        let min_len: i64 = len.floor() as i64;
        let max_len: i64 = len.ceil() as i64;

        let base_path = Path::new(&config.data_path);

        let mut inputs: Vec<Tensor> = Vec::new();
        let mut all_labels: Vec<Tensor> = Vec::new();

        for batch in 0..min_len {
            let mut batch_acc: Vec<Tensor> = Vec::new();
            let mut labels: Vec<Tensor> = Vec::new();
            for image in 0..batch_size {
                let i: usize = (batch * batch_size + image).try_into().unwrap();
                let item = &data[i];
                batch_acc.push(
                    tch::vision::image::load(base_path.join(&item.path))
                        .ok()
                        .unwrap(),
                );

                if item.class >= 0 {
                    let t = tch::Tensor::from_slice(&[item.class])
                        .onehot(classes_len.try_into().unwrap());
                    labels.push(t);
                } else {
                    labels.push(tch::Tensor::zeros(
                        (classes_len),
                        (tch::Kind::Float, tch::Device::Cpu),
                    ))
                }
            }
            inputs.push(tch::Tensor::cat(&batch_acc[0..], 0));
            all_labels.push(tch::Tensor::cat(&labels[0..], 0));
        }

        if min_len != max_len {
            let mut batch_acc: Vec<Tensor> = Vec::new();
            let mut labels: Vec<Tensor> = Vec::new();
            for image in 0..(data.len() - (batch_size * min_len) as usize) {
                let i: usize = (min_len * batch_size + (image as i64)) as usize;
                let item = &data[i];
                batch_acc.push(
                    tch::vision::image::load(base_path.join(&item.path))
                        .ok()
                        .unwrap(),
                );

                if item.class >= 0 {
                    let t = tch::Tensor::from_slice(&[item.class]).onehot(classes_len);
                    labels.push(t);
                } else {
                    labels.push(tch::Tensor::zeros(
                        classes_len,
                        (tch::Kind::Float, tch::Device::Cpu),
                    ))
                }
            }
            inputs.push(tch::Tensor::cat(&batch_acc[0..], 0));
            all_labels.push(tch::Tensor::cat(&labels[0..], 0));
        }

        return DataLoader {
            batch_size,
            inputs,
            labels: all_labels,
            len: max_len as usize,
            pos: 0,
        };
    }

    pub fn restart(self: &mut DataLoader) {
        self.pos = 0;
    }

    pub fn next(self: &mut DataLoader) -> Option<(Tensor, Tensor)> {
        if self.pos >= self.len {
            return None;
        }
        let input = self.inputs[self.pos].empty_like();
        self.inputs[self.pos] = self.inputs[self.pos].clone(&input);
        let label = self.labels[self.pos].empty_like();
        self.labels[self.pos] = self.labels[self.pos].clone(&label);

        return Some((input, label));
    }
}
