mod dataloader;
mod model;
mod settings;
mod tasks;
mod training;
mod types;

use crate::settings::*;
use crate::tasks::{fail_task, Task, TaskType};
use crate::training::handle_train;
use anyhow::{bail, Result};
use reqwest::StatusCode;
use serde_json::json;
use std::{fs, process::exit, sync::Arc, time::Duration};

enum ResultAlive {
    Ok,
    Error,
    NotInit,
}

async fn send_keep_alive_message(
    config: Arc<ConfigFile>,
    runner_data: Arc<RunnerData>,
) -> ResultAlive {
    let client = reqwest::Client::new();

    let to_send = json!({
        "id": runner_data.id,
    });

    let resp = client
        .post(format!("{}/tasks/runner/beat", config.hostname))
        .header("token", &config.token)
        .body(to_send.to_string())
        .send()
        .await;

    if resp.is_err() {
        return ResultAlive::Error;
    }

    let resp = resp.ok();

    if resp.is_none() {
        return ResultAlive::Error;
    }

    let resp = resp.unwrap();

    // TODO see if the message is related to not being inited
    if resp.status() != 200 {
        println!("Could not connect with the status");
        return ResultAlive::Error;
    }

    ResultAlive::Ok
}

async fn keep_alive(config: Arc<ConfigFile>, runner_data: Arc<RunnerData>) -> Result<()> {
    let mut failed = 0;
    loop {
        match send_keep_alive_message(config.clone(), runner_data.clone()).await {
            ResultAlive::Error => failed += 1,
            ResultAlive::NotInit => {
                println!("Runner not inited! Restarting!");
                exit(1)
            }
            ResultAlive::Ok => failed = 0,
        }

        // TODO move to config
        if failed > 20 {
            println!("Failed to connect to API! More than 20 times in a row stoping");
            exit(1)
        }

        tokio::time::sleep(Duration::from_secs(1)).await;
    }
}

async fn handle_task(
    task: Task,
    config: Arc<ConfigFile>,
    runner_data: Arc<RunnerData>,
) -> Result<()> {
    let res = match task.task_type {
        TaskType::Training => handle_train(&task, config.clone(), runner_data.clone()).await,
        _ => {
            println!("Do not know how to handle this task #{:?}", task);
            bail!("Failed")
        }
    };

    if res.is_err() {
        println!("task failed #{:?}", res);
        fail_task(
            &task,
            config,
            runner_data,
            "Do not know how to handle this kind of task",
        )
        .await?
    }

    Ok(())
}

#[tokio::main]
async fn main() -> Result<()> {
    // Load config file
    let config_data = fs::read_to_string("./config.toml")?;
    let mut config: ConfigFile = toml::from_str(&config_data)?;

    let client = reqwest::Client::new();
    if config.config_path == None {
        config.config_path = Some(String::from("./data.toml"))
    }

    let runner_data: RunnerData = load_runner_data(&config).await?;

    let to_send = json!({
        "id": runner_data.id,
    });

    // Inform the server that the runner is available
    let resp = client
        .post(format!("{}/tasks/runner/init", config.hostname))
        .header("token", &config.token)
        .body(to_send.to_string())
        .send()
        .await?;

    if resp.status() != 200 {
        println!(
            "Could not connect with the api: status {} body {}",
            resp.status(),
            resp.text().await?
        );
        return Ok(());
    }

    let res = resp.json::<String>().await?;
    if res != "Ok" {
        print!("Unexpected problem: {}", res);
        return Ok(());
    }

    let config = Arc::new(config);
    let runner_data = Arc::new(runner_data);

    let config_alive = config.clone();
    let runner_data_alive = runner_data.clone();
    std::thread::spawn(move || keep_alive(config_alive.clone(), runner_data_alive.clone()));

    println!("Started main loop");
    loop {
        //TODO move time to config
        tokio::time::sleep(Duration::from_secs(1)).await;

        let to_send = json!({ "id": runner_data.id });

        let resp = client
            .post(format!("{}/tasks/runner/active", config.hostname))
            .header("token", &config.token)
            .body(to_send.to_string())
            .send()
            .await;

        if resp.is_err() || resp.as_ref().ok().is_none() {
            println!("Failed to get info from server {:?}", resp);
            continue;
        }

        let resp = resp?;

        match resp.status() {
            // No active task
            StatusCode::NOT_FOUND => (),
            StatusCode::OK => {
                println!("Found task!");

                let task: Result<Task, reqwest::Error> = resp.json().await;
                if task.is_err() || task.as_ref().ok().is_none() {
                    println!("Failed to resolve the json {:?}", task);
                    continue;
                }

                let task = task?;

                let res = handle_task(task, config.clone(), runner_data.clone()).await;

                if res.is_err() || res.as_ref().ok().is_none() {
                    println!("Failed to run the task");
                }

                _ = res;
                ()
            }
            _ => {
                println!("Unexpected error #{:?}", resp);
                exit(1)
            }
        }
    }
}
