use anyhow::bail;
use serde::{Deserialize, Serialize};
use serde_repr::{Deserialize_repr, Serialize_repr};
use tch::{
    nn::{self, Module},
    Device,
};

#[derive(Debug)]
pub struct Model {
    pub vs: nn::VarStore,
    pub seq: nn::Sequential,
    pub layers: Vec<Layer>,
}

#[derive(Debug, Clone, Copy, Serialize_repr, Deserialize_repr)]
#[repr(i8)]
pub enum LayerType {
    Input = 1,
    Dense = 2,
    Flatten = 3,
    SimpleBlock = 4,
}

#[derive(Debug, Serialize, Deserialize)]
pub struct Layer {
    pub id: String,
    pub definition_id: String,
    pub layer_order: String,
    pub layer_type: LayerType,
    pub shape: String,
    pub exp_type: String,
}

#[derive(Debug, Serialize, Deserialize)]
pub struct DataPoint {
    pub class: i64,
    pub path: String,
}

pub fn build_model(layers: Vec<Layer>, last_linear_size: i64, add_sigmoid: bool) -> Model {
    let vs = nn::VarStore::new(Device::Cpu);

    let mut seq = nn::seq();

    let mut last_linear_size = last_linear_size;
    let mut last_linear_conv: Vec<i64> = Vec::new();

    for layer in layers.iter() {
        match layer.layer_type {
            LayerType::Input => {
                last_linear_conv = serde_json::from_str(&layer.shape).ok().unwrap();
                println!("Layer: Input, In: {:?}", last_linear_conv);
            }
            LayerType::Dense => {
                let shape: Vec<i64> = serde_json::from_str(&layer.shape).ok().unwrap();
                println!("Layer: Dense, In: {}, Out: {}", last_linear_size, shape[0]);
                seq = seq
                    .add(nn::linear(
                        &vs.root(),
                        last_linear_size,
                        shape[0],
                        Default::default(),
                    ))
                    .add_fn(|xs| xs.relu());
                last_linear_size = shape[0];
            }
            LayerType::Flatten => {
                seq = seq.add_fn(|xs| xs.flatten(1, -1));
                last_linear_size = 1;
                for i in &last_linear_conv {
                    last_linear_size *= i;
                }
                println!(
                    "Layer: flatten, In: {:?}, Out: {}",
                    last_linear_conv, last_linear_size
                )
            }
            LayerType::SimpleBlock => {
                panic!("DO not create Simple blocks yet");
                let new_last_linear_conv =
                    vec![128, last_linear_conv[1] / 2, last_linear_conv[2] / 2];
                println!(
                    "Layer: block, In: {:?}, Put: {:?}",
                    last_linear_conv, new_last_linear_conv,
                );
                //TODO
                //m_layers = append(m_layers, NewSimpleBlock(vs, lastLinearConv[0]))
                last_linear_conv = new_last_linear_conv;
            }
        }
    }

    if add_sigmoid {
        seq = seq.add_fn(|xs| xs.sigmoid());
    }

    return Model { vs, layers, seq };
}
